/*
 * Minimal Agar-GUI application (http://libagar.org).
 */

#include <agar/core.h>
#include <agar/gui.h>

int
main(int argc, char *argv[])
{
	AG_Window *win;

	if (AG_InitCore(NULL, 0) == -1 ||
	    AG_InitGraphics(0) == -1)
		return (1);

	win = AG_WindowNew(AG_WINDOW_MAIN);
	AG_LabelNew(win, 0, "Hello, world!");
	AG_WindowShow(win);

	AG_EventLoop();
	return (0);
}
